from __future__ import annotations

import tlc
from ultralytics.utils import LOGGER

from tlc_ultralytics.constants import TLC_COLORSTR


def reduce_embeddings(
    run: tlc.Run,
    method: str,
    n_components: int,
    foreign_table_url: tlc.Url | None = None,
    reducer_args: dict | None = None,
):
    """Reduce image embeddings by a foreign table URL."""
    if foreign_table_url is None:
        foreign_table_url = tlc.Url(tlc.active_run().constants["inputs"][0]["input_table_url"]).to_absolute(
            tlc.active_run().url
        )

    LOGGER.info(
        TLC_COLORSTR + f"Reducing image embeddings to {n_components}D with {method}, this may take a few minutes..."
    )
    run.reduce_embeddings_by_foreign_table_url(
        foreign_table_url=foreign_table_url,
        method=method,
        n_components=n_components,
        **(reducer_args or {}),
    )



from __future__ import annotations

from typing import Any, Callable, Literal

import numpy as np
import tlc
from tlc.core.builtins.types.bounding_box import CenteredXYWHBoundingBox
from ultralytics.data.dataset import YOLODataset
from ultralytics.data.utils import check_file_speeds, segments2boxes
from ultralytics.utils import LOGGER, colorstr

from tlc_ultralytics.engine.dataset import TLCDatasetMixin

SegmentType = Literal["absolute", "relative"]


class IdentityDict(dict):
    def __missing__(self, key):
        return key


class TLCYOLODataset:
    """Factory class for creating task-specific 3LC YOLO datasets."""

    def __new__(
        cls,
        table,
        data=None,
        task="detect",
        exclude_zero=False,
        class_map=None,
        image_column_name=None,
        label_column_name=None,
        **kwargs,
    ):
        """Create a new dataset instance of the appropriate type.

        :param table: The 3LC table containing the dataset
        :param data: Optional data parameter for YOLODataset
        :param task: Either "segment" or "detect"
        :param exclude_zero: Whether to exclude zero-class annotations
        :param class_map: Optional mapping from original class indices to new ones
        :param image_column_name: Name of the image column in the table
        :param label_column_name: Name of the label column in the table
        :param **kwargs: Additional arguments passed to the dataset constructor
        """
        if task == "detect":
            return TLCYOLODetectionDataset(
                table=table,
                data=data,
                exclude_zero=exclude_zero,
                class_map=class_map,
                image_column_name=image_column_name,
                label_column_name=label_column_name,
                **kwargs,
            )
        elif task == "segment":
            return TLCYOLOSegmentationDataset(
                table=table,
                data=data,
                exclude_zero=exclude_zero,
                class_map=class_map,
                image_column_name=image_column_name,
                label_column_name=label_column_name,
                **kwargs,
            )
        else:
            raise ValueError(f"Unsupported task: {task} for TLCYOLODataset. Only 'segment' and 'detect' are supported.")


class BaseTLCYOLODataset(TLCDatasetMixin, YOLODataset):
    """Base class for 3LC YOLO datasets.

    This class provides common functionality for any detection task.
    Task-specific functionality should be implemented in subclasses.
    """

    def __init__(
        self,
        table,
        data=None,
        exclude_zero=False,
        class_map=None,
        image_column_name=None,
        label_column_name=None,
        **kwargs,
    ):
        """Initialize the base dataset.

        :param table: The 3LC table containing the dataset
        :param data: Optional data parameter for YOLODataset
        :param exclude_zero: Whether to exclude zero-class annotations
        :param class_map: Optional mapping from original class indices to new ones
        :param image_column_name: Name of the image column in the table
        :param label_column_name: Name of the label column in the table
        """
        self.table = table
        self._exclude_zero = exclude_zero
        self._class_map = class_map if class_map is not None else IdentityDict()
        self._image_column_name = image_column_name
        self._label_column_name = label_column_name

        super().__init__(table, data=data, **kwargs)
        self._post_init()

    def get_img_files(self, _):
        """Images are read in `get_labels` to avoid two loops, return empty list here."""
        im_files, labels = self._get_rows_from_table()
        check_file_speeds(im_files, prefix=colorstr(self.prefix + ":") + " ")
        self.labels = labels
        self.im_files = im_files
        return self.im_files

    def get_labels(self):
        """Get the labels from the table."""
        return self.labels

    def _index_to_example_id(self, index: int) -> int:
        """Get the example id for the given index."""
        return self.labels[index]["example_id"]

    def _get_label_from_row(self, im_file: str, row: Any, example_id: int) -> dict[str, Any]:
        """Get the label for a row in the appropriate format.

        This method should be implemented by subclasses.
        """
        raise NotImplementedError("Subclasses must implement _get_label_from_row")


class TLCYOLODetectionDataset(BaseTLCYOLODataset):
    """3LC YOLO dataset for object detection."""

    def __init__(
        self,
        table,
        data=None,
        exclude_zero=False,
        class_map=None,
        image_column_name=None,
        label_column_name=None,
        **kwargs,
    ):
        """Initialize the detection dataset.

        :param table: The 3LC table containing the dataset
        :param data: Optional data parameter for YOLODataset
        :param exclude_zero: Whether to exclude zero-class annotations
        :param class_map: Optional mapping from original class indices to new ones
        :param image_column_name: Name of the image column in the table
        :param label_column_name: Name of the label column in the table
        """
        self._detection_factory: Callable[[list[float]], tlc.BoundingBox] = self._get_detection_factory(
            table, label_column_name
        )

        super().__init__(
            table,
            data=data,
            task="detect",
            exclude_zero=exclude_zero,
            class_map=class_map,
            image_column_name=image_column_name,
            label_column_name=label_column_name,
            **kwargs,
        )

    def _get_detection_factory(
        self, table: tlc.Table, label_column_name: str
    ) -> Callable[[list[float]], tlc.BoundingBox]:
        """Infer the bounding box factory from the table schema.

        :param table: The 3LC table containing the dataset
        :param label_column_name: The name of the label column in the table
        :returns: A factory function that creates bounding boxes from coordinates
        """
        column_name, instances_name, _ = label_column_name.split(".")

        try:
            factory = tlc.BoundingBox.from_schema(table.rows_schema.values[column_name].values[instances_name])
        except Exception as e:
            raise ValueError(f"Table {table.url} is not a detection table: {e}") from None

        return factory

    def _get_label_from_row(self, im_file: str, row: Any, example_id: int) -> dict[str, Any]:
        """Get the detection label for a row."""
        return tlc_table_row_to_yolo_label(
            row,
            self._detection_factory,
            self._class_map,
            im_file,
            label_column_name=self._label_column_name,
            example_id=example_id,
        )


class TLCYOLOSegmentationDataset(BaseTLCYOLODataset):
    """3LC YOLO dataset for instance segmentation."""

    def __init__(
        self,
        table,
        data=None,
        exclude_zero=False,
        class_map=None,
        image_column_name=None,
        label_column_name=None,
        **kwargs,
    ):
        """Initialize the segmentation dataset.

        :param table: The 3LC table containing the dataset
        :param data: Optional data parameter for YOLODataset
        :param exclude_zero: Whether to exclude zero-class annotations
        :param class_map: Optional mapping from original class indices to new ones
        :param image_column_name: Name of the image column in the table
        :param label_column_name: Name of the label column in the table
        """
        self._segment_type: SegmentType = self._get_segment_type(table, label_column_name)
        super().__init__(
            table,
            data=data,
            task="segment",
            exclude_zero=exclude_zero,
            class_map=class_map,
            image_column_name=image_column_name,
            label_column_name=label_column_name,
            **kwargs,
        )

    def _get_segment_type(self, table: tlc.Table, label_column_name: str) -> SegmentType:
        """Verify the table format and check if the polygons are relative.

        :param table: The 3LC table containing the dataset
        :param label_column_name: The name of the label column in the table
        :returns: The segment type ("absolute" or "relative")
        """
        column_name, instances_name, _ = label_column_name.split(".")

        try:
            rles_schema_value = table.rows_schema.values[column_name].values["rles"].value
            segment_type = "relative" if getattr(rles_schema_value, "polygons_are_relative", False) else "absolute"
        except Exception as e:
            raise ValueError(f"Table {table.url} is not a segmentation table: {e}") from None

        return segment_type

    def _normalize_segments(self, segments: list[np.ndarray], width: int, height: int) -> list[np.ndarray]:
        """Normalize segments to relative coordinates if they are absolute.

        :param segments: List of segment coordinates
        :param width: Image width
        :param height: Image height
        :return: Normalized segments
        """
        if self._segment_type == "absolute":
            return segments / np.array([width, height])
        return segments

    def _get_label_from_row(self, im_file: str, row: Any, example_id: int) -> dict[str, Any]:
        """Get the segmentation label for a row."""
        column_name, instances_name, label_key = self._label_column_name.split(".")

        # Use sample view to get polygons, row is row view
        sample = self.table[example_id]

        segmentations = sample[column_name]
        height, width = segmentations[tlc.IMAGE_HEIGHT], segmentations[tlc.IMAGE_WIDTH]
        classes = []
        segments = []

        for i, (category, polygon) in enumerate(
            zip(
                segmentations[instances_name][label_key],
                segmentations[tlc.POLYGONS],
            )
        ):
            # Handle polygons with zero area
            if len(polygon) < 6:
                LOGGER.warning(f"Polygon {i} in row {example_id} has fewer than 6 points and will be ignored.")
                continue

            classes.append(self._class_map[category])
            row_segments = np.array(polygon, dtype=np.float32).reshape(-1, 2)
            segments.append(self._normalize_segments(row_segments, width, height))

        # Compute bounding boxes from segments
        if segments:
            bboxes = segments2boxes(segments)
        else:
            bboxes = np.zeros((0, 4), dtype=np.float32)

        return {
            "im_file": im_file,
            "shape": (height, width),  # format: (height, width)
            "cls": np.array(classes).reshape(-1, 1),
            "bboxes": bboxes,
            "segments": segments,
            "keypoints": None,
            "normalized": True,
            "bbox_format": "xywh",
            "example_id": example_id,
        }


def convert_to_xywh(bbox: tlc.BoundingBox, image_width: int, image_height: int) -> CenteredXYWHBoundingBox:
    if isinstance(bbox, CenteredXYWHBoundingBox):
        return bbox
    else:
        return CenteredXYWHBoundingBox.from_top_left_xywh(bbox.to_top_left_xywh().normalize(image_width, image_height))


def unpack_box(
    bbox: dict[str, int | float],
    table_format: Callable[[list[float]], tlc.BoundingBox],
    image_width: int,
    image_height: int,
    label_key: str,
) -> tuple[int, list[float]]:
    coordinates = [bbox[tlc.X0], bbox[tlc.Y0], bbox[tlc.X1], bbox[tlc.Y1]]
    return bbox[label_key], convert_to_xywh(table_format(coordinates), image_width, image_height)


def unpack_boxes(
    bboxes: list[dict[str, int | float]],
    class_map: dict[int, int],
    table_format: Callable[[list[float]], tlc.BoundingBox],
    image_width: int,
    image_height: int,
    label_key: str,
) -> tuple[np.ndarray, np.ndarray]:
    classes_list, boxes_list = [], []
    for bbox in bboxes:
        _class, box = unpack_box(bbox, table_format, image_width, image_height, label_key)

        # Ignore boxes with non-positive width or height
        if box[2] > 0 and box[3] > 0:
            classes_list.append(class_map[_class])
            boxes_list.append(box)

    # Convert to np array
    boxes = np.array(boxes_list, ndmin=2, dtype=np.float32)
    if len(boxes_list) == 0:
        boxes = boxes.reshape(0, 4)

    classes = np.array(classes_list, dtype=np.float32).reshape((-1, 1))
    assert classes.shape == (boxes.shape[0], 1)
    return classes, boxes


def tlc_table_row_to_yolo_label(
    row,
    detection_factory: Callable[[list[float]], tlc.BoundingBox],
    class_map: dict[int, int],
    im_file: str,
    label_column_name: str,
    example_id: int,
) -> dict[str, Any]:
    """Convert a table row from a 3lc Table to a Ultralytics YOLO label dict.

    :param row: The table row to convert
    :param detection_factory: Factory function to create bounding boxes
    :param class_map: A dictionary mapping 3lc class labels to contiguous class labels
    :param im_file: The path to the image file of the row
    :param label_column_name: The name of the label column in the table
    :returns: A dictionary containing the Ultralytics YOLO label information
    """
    bounding_boxes_column_key, bounding_boxes_list_key, label_key = label_column_name.split(".")

    classes, bboxes = unpack_boxes(
        row[bounding_boxes_column_key][bounding_boxes_list_key],
        class_map,
        detection_factory,
        row[bounding_boxes_column_key][tlc.IMAGE_WIDTH],
        row[bounding_boxes_column_key][tlc.IMAGE_HEIGHT],
        label_key,
    )

    return {
        "im_file": im_file,
        "shape": (
            row[bounding_boxes_column_key][tlc.IMAGE_HEIGHT],
            row[bounding_boxes_column_key][tlc.IMAGE_WIDTH],
        ),  # format: (height, width)
        "cls": classes,
        "bboxes": bboxes,
        "segments": [],
        "keypoints": None,
        "normalized": True,
        "bbox_format": "xywh",
        "example_id": example_id,
    }



from __future__ import annotations

from collections.abc import Callable, Iterable
from pathlib import Path

import tlc
import yaml
from ultralytics.utils import LOGGER, colorstr

from tlc_ultralytics.constants import TLC_COLORSTR, TLC_PREFIX


def check_tlc_dataset(  # noqa: C901
    data: str,
    tables: dict[str, tlc.Table | tlc.Url | str] | None,
    image_column_name: str,
    label_column_name: str,
    dataset_checker: Callable[[str], dict[str, object]] | None = None,
    table_creator: Callable[[str, dict[str, object], str, str, str, str, str], tlc.Table] | None = None,
    table_checker: Callable[[str, tlc.Table], bool] | None = None,
    project_name: str | None = None,
    splits: Iterable[str] | None = None,
) -> dict[str, tlc.Table | dict[float, str] | int]:
    """Get or create tables for YOLO datasets. data is ignored when tables is provided.

    :param data: Path to a dataset
    :param tables: Dictionary of tables, if already created
    :param image_column_name: Name of the column containing image paths
    :param label_column_name: Name of the column containing labels
    :param dataset_checker: Function to check the dataset (yolo implementation, download and checks)
    :param table_creator: Function to create the tables for the YOLO dataset
    :param table_checker: Function to check that a table is compatible with the current task
    :param project_name: Name of the project
    :param splits: List of splits to parse.
    :return: Dictionary of tables and class names
    """

    if not tables and data is not None and data.endswith(".ndjson"):
        raise ValueError(
            "NDJson datasets are not yet supported in the YOLO integration. Create a tlc.Table from the ",
            "data or convert it to a YOLO dataset and use `tlc.Table.from_yolo`. The NDJson format will be ",
            "supported in the future.",
        )

    # If the data starts with the 3LC prefix, parse the YAML file and populate `tables`
    has_prefix = False
    if tables is None and isinstance(data, str) and data.startswith(TLC_PREFIX):
        has_prefix = True
        LOGGER.info(f"{TLC_COLORSTR}Parsing 3LC YAML file data={data} and populating tables")
        tables = parse_3lc_yaml_file(data)

    if tables is None:
        tables = {}

        data_dict = dataset_checker(data)

        # Get or create tables
        splits = splits or ("train", "val", "test", "minival")

        for key in splits:
            if data_dict.get(key):
                name = Path(data).stem
                dataset_name = f"{name}-{key}"
                table_name = "initial"

                if project_name is None:
                    project_name = f"{name}-YOLO"

                try:
                    table = table_creator(
                        key,
                        data_dict,
                        image_column_name=image_column_name,
                        label_column_name=label_column_name,
                        project_name=project_name,
                        dataset_name=dataset_name,
                        table_name=table_name,
                    )

                    # Get the latest version when inferring
                    tables[key] = table.latest()

                    if tables[key] != table:
                        LOGGER.info(
                            f"{colorstr(key)}: Using latest version of table from {data}: "
                            f"{table.url} -> {tables[key].url}"
                        )
                    else:
                        LOGGER.info(f"{colorstr(key)}: Using initial version of table from {data}: {tables[key].url}")

                except Exception as e:
                    LOGGER.warning(
                        f"{colorstr(key)}: Failed to read or create table for split {key} from {data}: {e!s}"
                    )

    else:
        # LOGGER.info(f"{TLC_COLORSTR}Using data directly from tables")
        _check_tables(tables)

        for key, table in tables.items():
            if splits is not None and key not in splits:
                continue

            if isinstance(table, (str, Path, tlc.Url)):
                try:
                    table_url = tlc.Url(table)
                    tables[key] = tlc.Table.from_url(table_url)
                except Exception as e:
                    raise ValueError(
                        f"Error loading table from {table} for split '{key}' provided through `tables`."
                    ) from e
            elif isinstance(table, tlc.Table):
                tables[key] = table
            else:
                msg = (
                    f"Invalid type {type(table)} for split {key} provided through `tables`."
                    "Must be a tlc.Table object or a location (string, pathlib.Path or tlc.Url) of a tlc.Table."
                )

                raise ValueError(msg)

            # Check that the table is compatible with the current task
            if table_checker is not None:
                table_checker(tables[key], image_column_name, label_column_name)

            source = "3LC YAML file" if has_prefix else "provided tables"
            LOGGER.info(f"{colorstr(key)}: Using table {tables[key].url} from {source}")

    first_split = next(iter(tables.keys()))

    names = tables[first_split].get_simple_value_map(label_column_name)
    value_map = tables[first_split].get_value_map(label_column_name)

    if names is None:
        raise ValueError(f"Failed to get value map for table with Url: {tables[first_split].url}")

    for split, split_table in tables.items():
        split_names = split_table.get_simple_value_map(label_column_name)

        if split_names is None:
            raise ValueError(f"Failed to get value map for table with Url: {tables[split].url}")

        if split_names != names:
            first_items = set(names.items())
            split_items = set(split_names.items())

            only_in_first = first_items - split_items
            only_in_split = split_items - first_items

            messages = []

            if only_in_first:
                dict_str = "{" + ", ".join(f"{k}: '{v}'" for k, v in only_in_first) + "}"
                messages.append(f"'{first_split}' has categories that '{split}' does not: {dict_str}")
            if only_in_split:
                dict_str = "{" + ", ".join(f"{k}: '{v}'" for k, v in only_in_split) + "}"
                messages.append(f"'{split}' has categories that '{first_split}' does not: {dict_str}")

            error_msg = "All splits must have the same categories, but " + " and ".join(messages)

            raise ValueError(error_msg)

    # Map name indices to 0, 1, ..., n-1
    names_yolo = dict(enumerate(names.values()))
    range_to_3lc_class = dict(enumerate(names))

    return {
        **tables,
        "names": names_yolo,
        "names_3lc": value_map,
        "nc": len(names),
        "range_to_3lc_class": range_to_3lc_class,
        "3lc_class_to_range": {v: k for k, v in range_to_3lc_class.items()},
        "channels": 3,  # TODO(Frederik): Read out channels from appropriate place and populate here
    }


def parse_3lc_yaml_file(data_file: str) -> dict[str, tlc.Table]:
    """Parse a 3LC YAML file and return the corresponding tables.

    :param data_file: The path to the 3LC YAML file.
    :returns: The tables pointed to by the YAML file.
    """
    # Read the YAML file, removing the prefix
    if not (data_file_url := tlc.Url(data_file.replace(TLC_PREFIX, ""))).exists():
        raise FileNotFoundError(f"Could not find YAML file {data_file_url}")

    data_config = yaml.safe_load(data_file_url.read())

    path = data_config.get("path")
    splits = [key for key in data_config if key != "path"]

    tables = {}
    for split in splits:
        # Handle :latest at the end
        if data_config[split].endswith(":latest"):
            latest = True
            split_path = data_config[split][: -len(":latest")]
        else:
            latest = False
            split_path = data_config[split]

        if split_path.startswith("./"):
            LOGGER.debug(f"{TLC_COLORSTR}{split} split path starts with './', removing it.")
            split_path = split_path[2:]

        table_url = tlc.Url(path) / split_path if path else tlc.Url(split_path)

        table = tlc.Table.from_url(table_url)

        if latest:
            table = table.latest()

        tables[split] = table

    return tables


def _check_tables(tables: object):
    if not isinstance(tables, dict):
        msg = f"When providing tables directly, they must be a dictionary, but got type {type(tables)}."
        raise ValueError(msg)

    for key, table in tables.items():
        if not isinstance(table, (str, Path, tlc.Url, tlc.Table)):
            msg = (
                "When providing tables directly, they must be a tlc.Table or a URL to a tlc.Table. ",
                f"Got {type(table)} for split {key}.",
            )
            raise ValueError(msg)



from __future__ import annotations

import torch
from ultralytics.utils.loss import BboxLoss, v8DetectionLoss
from ultralytics.utils.metrics import bbox_iou
from ultralytics.utils.tal import bbox2dist, make_anchors


class UnreducedBboxLoss(BboxLoss):
    """BboxLoss that returns unreduced losses for per-sample computation."""

    def forward(
        self,
        pred_dist: torch.Tensor,
        pred_bboxes: torch.Tensor,
        anchor_points: torch.Tensor,
        target_bboxes: torch.Tensor,
        target_scores: torch.Tensor,
        target_scores_sum: float,
        fg_mask: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Forward pass computing IoU and DFL losses.

        :param pred_dist: Predicted distribution tensor
        :param pred_bboxes: Predicted bounding boxes
        :param anchor_points: Anchor points for the predictions
        :param target_bboxes: Target bounding boxes
        :param target_scores: Target scores
        :param target_scores_sum: Sum of target scores
        :param fg_mask: Foreground mask
        :return: Tuple of (iou_loss, dfl_loss) tensors
        """
        weight = target_scores.sum(-1)[fg_mask].unsqueeze(-1)
        iou = bbox_iou(pred_bboxes[fg_mask], target_bboxes[fg_mask], xywh=False, CIoU=True)
        loss_iou = (1.0 - iou) * weight

        if self.dfl_loss:
            target_ltrb = bbox2dist(anchor_points, target_bboxes, self.dfl_loss.reg_max - 1)
            loss_dfl = (
                self.dfl_loss(
                    pred_dist[fg_mask].view(-1, self.dfl_loss.reg_max),
                    target_ltrb[fg_mask],
                )
                * weight
            )
        else:
            loss_dfl = torch.zeros_like(loss_iou)

        assert loss_iou.shape == loss_dfl.shape, f"IoU Loss shape {loss_iou.shape} != DFL Loss shape {loss_dfl.shape}"

        return loss_iou, loss_dfl


class v8UnreducedDetectionLoss(v8DetectionLoss):
    """v8DetectionLoss that returns unreduced losses for per-sample computation."""

    def __init__(self, model, tal_topk: int = 10, training: bool = False):
        """Initialize the unreduced detection loss.

        :param model: The YOLO model
        :param tal_topk: Top-k for TAL assignment
        :param training: Whether in training mode
        """
        super().__init__(model, tal_topk=tal_topk)

        m = model.model[-1]  # Detect() module
        self.bbox_loss = UnreducedBboxLoss(m.reg_max)
        self.training = training

    def __call__(self, preds, batch) -> dict[str, torch.Tensor]:
        """Calculate unreduced losses for box, cls and dfl.

        :param preds: Model predictions
        :param batch: Batch data
        :return: Dictionary containing unreduced losses
        """
        feats = preds[1] if isinstance(preds, (list, tuple)) else preds
        pred_distri, pred_scores = torch.cat([xi.view(feats[0].shape[0], self.no, -1) for xi in feats], 2).split(
            (self.reg_max * 4, self.nc), 1
        )

        pred_scores = pred_scores.permute(0, 2, 1).contiguous()
        pred_distri = pred_distri.permute(0, 2, 1).contiguous()

        dtype = pred_scores.dtype
        batch_size = pred_scores.shape[0]
        imgsz = torch.tensor(feats[0].shape[2:], device=self.device, dtype=dtype) * self.stride[0]
        anchor_points, stride_tensor = make_anchors(feats, self.stride, 0.5)

        # Targets
        targets = torch.cat(
            (batch["batch_idx"].view(-1, 1), batch["cls"].view(-1, 1), batch["bboxes"]),
            1,
        )
        targets = self.preprocess(targets.to(self.device), batch_size, scale_tensor=imgsz[[1, 0, 1, 0]])
        gt_labels, gt_bboxes = targets.split((1, 4), 2)  # cls, xyxy
        mask_gt = gt_bboxes.sum(2, keepdim=True).gt_(0.0)

        # Pboxes
        pred_bboxes = self.bbox_decode(anchor_points, pred_distri)  # xyxy, (b, h*w, 4)

        _, target_bboxes, target_scores, fg_mask, _ = self.assigner(
            pred_scores.detach().sigmoid(),
            (pred_bboxes.detach() * stride_tensor).type(gt_bboxes.dtype),
            anchor_points * stride_tensor,
            gt_labels,
            gt_bboxes,
            mask_gt,
        )

        target_scores_sum = max(target_scores.sum(), 1)

        # Cls loss
        cls_loss = self.bce(pred_scores, target_scores.to(dtype)).sum(dim=2)

        # Bbox loss
        box_loss_full = torch.zeros_like(cls_loss)
        dfl_loss_full = torch.zeros_like(cls_loss)
        if fg_mask.sum():
            target_bboxes /= stride_tensor
            box_loss, dfl_loss = self.bbox_loss(
                pred_distri,
                pred_bboxes,
                anchor_points,
                target_bboxes,
                target_scores,
                target_scores_sum,
                fg_mask,
            )
            box_loss_full[fg_mask] = box_loss.to(cls_loss.dtype).squeeze()
            dfl_loss_full[fg_mask] = dfl_loss.to(cls_loss.dtype).squeeze()

        losses = {
            "cls_loss": cls_loss,
            "box_loss": box_loss_full,
            "dfl_loss": dfl_loss_full,
        }

        if self.training:
            cls_weight = self.hyp.cls if hasattr(self.hyp, "cls") else self.hyp["cls"]
            box_weight = self.hyp.box if hasattr(self.hyp, "box") else self.hyp["box"]
            dfl_weight = self.hyp.dfl if hasattr(self.hyp, "dfl") else self.hyp["dfl"]
            losses["loss"] = cls_loss * cls_weight + box_loss_full * box_weight + dfl_loss_full * dfl_weight

        return losses



from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path

import tlc
from ultralytics.data.utils import check_det_dataset

from tlc_ultralytics.utils import check_tlc_dataset


def tlc_check_seg_dataset(
    data: str,
    tables: dict[str, tlc.Table | tlc.Url | Path | str] | None,
    image_column_name: str,
    label_column_name: str,
    project_name: str | None = None,
    splits: Iterable[str] | None = None,
) -> dict[str, tlc.Table | dict[float, str] | int]:
    return check_tlc_dataset(
        data,
        tables,
        image_column_name,
        label_column_name,
        dataset_checker=check_det_dataset,
        table_creator=get_or_create_seg_table,
        table_checker=check_seg_table,
        project_name=project_name,
        splits=splits,
    )


def get_or_create_seg_table(
    key: str,
    data_dict: dict[str, object],
    image_column_name: str,
    label_column_name: str,
    project_name: str,
    dataset_name: str,
    table_name: str,
) -> tlc.Table:
    return tlc.Table.from_yolo(
        dataset_yaml_file=data_dict["yaml_file"],
        split=key,
        override_split_path=data_dict[key],
        task="segment",
        project_name=project_name,
        dataset_name=dataset_name,
        table_name=table_name,
        if_exists="reuse",
        add_weight_column=True,
        description="Created with 3LC YOLO integration",
    )


def check_seg_table(table: tlc.Table, image_column_name: str, label_column_name: str) -> None:
    """Verify that the table is compatible with instance segmentation.

    :param table: The table to check.
    :param image_column_name: The name of the image column.
    :param label_column_name: The value path of the label.
    :raises ValueError: If the table is not compatible with instance segmentation.
    """
    row_schema = table.row_schema.values

    label_column_name = label_column_name.split(".")[0]

    # Check that the schema is compatible with instance segmentation
    try:
        # Schema checks
        assert image_column_name in row_schema, f"Image column '{image_column_name}' not found."
        assert label_column_name in row_schema, f"Label column '{label_column_name}' not found."

        assert hasattr(row_schema[label_column_name], "sample_type"), (
            f"Label column '{label_column_name}' does not have a sample type."
        )
        sample_type = tlc.SampleType.from_schema(row_schema[label_column_name])
        assert isinstance(sample_type, tlc.InstanceSegmentationPolygons), (
            f"Label column '{label_column_name}' does not have sample type InstanceSegmentationPolygons."
        )

    except AssertionError as e:
        msg = f"Schema validation failed for '{label_column_name}' column in table with URL {table.url}. {e!s}"
        raise ValueError(msg) from e

    # Check that the table data is compatible with its schema
    try:
        first_row = table[0]
        sample_type.ensure_sample_valid(first_row[label_column_name])
        assert image_column_name in first_row, (
            f"Image column {image_column_name} not found in table with URL {table.url}"
        )

    except (AssertionError, ValueError) as e:
        msg = f"Data validation failed for {label_column_name} column in table with URL {table.url}. {e!s}"
        raise ValueError(msg) from e



from __future__ import annotations

import weakref

import tlc
import torch
from ultralytics.models.yolo.detect import DetectionValidator
from ultralytics.utils import metrics, ops

from tlc_ultralytics.constants import (
    DETECTION_LABEL_COLUMN_NAME,
    IMAGE_COLUMN_NAME,
)
from tlc_ultralytics.detect.loss import v8UnreducedDetectionLoss
from tlc_ultralytics.detect.utils import (
    build_tlc_yolo_dataset,
    construct_bbox_struct,
    tlc_check_det_dataset,
    yolo_loss_schemas,
    yolo_predicted_bounding_box_schema,
)
from tlc_ultralytics.engine.validator import TLCValidatorMixin


class TLCDetectionValidator(TLCValidatorMixin, DetectionValidator):
    _default_image_column_name = IMAGE_COLUMN_NAME
    _default_label_column_name = DETECTION_LABEL_COLUMN_NAME

    def check_dataset(self, *args, **kwargs):
        return tlc_check_det_dataset(*args, **kwargs)

    def build_dataset(self, table, mode="val", batch=None):
        return build_tlc_yolo_dataset(
            self.args,
            table,
            batch,
            self.data,
            mode=mode,
            stride=self.stride,
            exclude_zero=self._settings.exclude_zero_weight_collection,
            class_map=self.data["3lc_class_to_range"],
            split=self.args.split,
            image_column_name=self._image_column_name,
            label_column_name=self._label_column_name,
        )

    def postprocess(self, preds):
        self._curr_raw_preds = preds if self._settings.collect_loss else None
        return super().postprocess(preds)

    def _get_metrics_schemas(self):
        loss_schemas = yolo_loss_schemas(training=self._training) if self._settings.collect_loss else {}

        return {
            tlc.PREDICTED_BOUNDING_BOXES: yolo_predicted_bounding_box_schema(self.data["names_3lc"]),
            **loss_schemas,
        }

    def _compute_3lc_metrics(self, preds, batch):
        losses = self.loss_fn(self._curr_raw_preds, batch) if self._settings.collect_loss else {}

        processed_predictions = self._process_detection_predictions(preds, batch)
        return {
            tlc.PREDICTED_BOUNDING_BOXES: processed_predictions,
            **{k: tensor.mean(dim=1).cpu().numpy() for k, tensor in losses.items()},
        }

    def _process_detection_predictions(self, batch_predictions, batch):
        batch_predicted_boxes = []

        for i, predictions in enumerate(batch_predictions):
            predicted_boxes, predicted_confidences, predicted_classes = (
                predictions["bboxes"].clone(),
                predictions["conf"].clone(),
                predictions["cls"].clone(),
            )
            height, width = batch["ori_shape"][i]

            # Handle case with no predictions
            if len(predictions) == 0:
                batch_predicted_boxes.append(
                    construct_bbox_struct(
                        [],
                        image_width=width,
                        image_height=height,
                    )
                )
                continue

            # Handle case with predictions
            else:
                # Filter out low confidence predictions
                mask = predicted_confidences > self._settings.conf_thres
                predicted_boxes = predicted_boxes[mask]
                predicted_confidences = predicted_confidences[mask].tolist()
                predicted_classes = predicted_classes[mask].tolist()

                # Compute IoUs
                pbatch = self._prepare_batch(i, batch)
                gt_boxes = pbatch["bboxes"].clone()
                if gt_boxes.shape[0]:
                    ious = metrics.box_iou(gt_boxes, predicted_boxes)  # IoU evaluated in xyxy format
                    box_ious = ious.max(dim=0)[0].cpu().tolist()
                else:
                    box_ious = [0.0] * predicted_boxes.shape[0]  # No predictions

                # Scale predicted boxes to original image size
                resized_shape = batch["resized_shape"][i]
                ori_shape = batch["ori_shape"][i]
                ratio_pad = batch["ratio_pad"][i]
                pred_scaled = ops.scale_boxes(resized_shape, predicted_boxes, ori_shape, ratio_pad)

                pred_xywh = ops.xyxy2xywhn(pred_scaled, w=width, h=height)

                annotations = []
                for pi in range(len(predicted_boxes)):
                    annotations.append(
                        {
                            "score": predicted_confidences[pi],
                            "category_id": self.data["range_to_3lc_class"][int(predicted_classes[pi])],
                            "bbox": pred_xywh[pi, :].cpu().tolist(),
                            "iou": box_ious[pi],
                        }
                    )

                batch_predicted_boxes.append(
                    construct_bbox_struct(
                        annotations,
                        image_width=width,
                        image_height=height,
                    )
                )

        return batch_predicted_boxes

    def _prepare_loss_fn(self, model):
        self.loss_fn = v8UnreducedDetectionLoss(
            model.model if hasattr(model.model, "model") else model,
            training=self._training,
        )

    def _add_embeddings_hook(self, model) -> int:
        if hasattr(model.model, "model"):
            model = model.model

        # Find index of the SPPF layer
        sppf_index = next((i for i, m in enumerate(model.model) if "SPPF" in m.type), -1)

        if sppf_index == -1:
            raise ValueError(
                "Image level embeddings can only be collected for detection models with a SPPF layer, "
                "but this model does not have one."
            )

        weak_self = weakref.ref(self)  # Avoid circular reference (self <-> hook_fn)

        def hook_fn(_module, _input, output):
            # Store embeddings
            self_ref = weak_self()
            flattened_output = torch.nn.functional.adaptive_avg_pool2d(output, (1, 1)).squeeze(-1).squeeze(-1)
            embeddings = flattened_output.detach().cpu().numpy()
            self_ref.embeddings = embeddings

        # Add forward hook to collect embeddings
        for i, module in enumerate(model.model):
            if i == sppf_index:
                self._hook_handles.append(module.register_forward_hook(hook_fn))

        activation_size = model.model[sppf_index]._modules["cv2"]._modules["conv"].out_channels
        return activation_size

    def _infer_batch_size(self, preds, batch) -> int:
        return len(batch["im_file"])



# ruff: noqa: E402

import random

import sentry_sdk

sentry_sdk.profiler.transaction_profiler.random = random.Random()

from importlib.metadata import PackageNotFoundError, version

from tlc_ultralytics.engine.model import TLCYOLO, YOLO
from tlc_ultralytics.settings import Settings

try:
    __version__ = version("3lc-ultralytics")
except PackageNotFoundError:
    __version__ = "unknown"

__all__ = [
    "TLCYOLO",
    "YOLO",
    "Settings",
]



from __future__ import annotations

from functools import partial

import ultralytics
from ultralytics.models import yolo

from tlc_ultralytics.classify.dataset import TLCClassificationDataset
from tlc_ultralytics.classify.utils import tlc_check_cls_dataset
from tlc_ultralytics.classify.validator import TLCClassificationValidator
from tlc_ultralytics.constants import (
    CLASSIFY_LABEL_COLUMN_NAME,
    IMAGE_COLUMN_NAME,
)
from tlc_ultralytics.engine.trainer import TLCTrainerMixin
from tlc_ultralytics.overrides import build_dataloader
from tlc_ultralytics.utils import create_sampler


class TLCClassificationTrainer(TLCTrainerMixin, yolo.classify.ClassificationTrainer):
    _default_image_column_name = IMAGE_COLUMN_NAME
    _default_label_column_name = CLASSIFY_LABEL_COLUMN_NAME

    def get_dataset(self):
        """Overrides the get_dataset method to get or create 3LC tables."""
        self.data = tlc_check_cls_dataset(
            self.args.data,
            self._tables,
            self._image_column_name,
            self._label_column_name,
            project_name=self._settings.project_name,
            splits=("train", "val"),
        )
        if "val" not in self.data:
            data_test = tlc_check_cls_dataset(
                self.args.data,
                self._tables,
                self._image_column_name,
                self._label_column_name,
                project_name=self._settings.project_name,
                splits=("test",),
            )
            self.data["test"] = data_test["test"]

        return self.data

    def build_dataset(self, table, mode="train", batch=None):
        exclude_zero = mode == "val" and self._settings.exclude_zero_weight_collection

        return TLCClassificationDataset(
            table,
            args=self.args,
            augment=mode == "train",
            prefix=mode,
            image_column_name=self._image_column_name,
            label_column_name=self._label_column_name,
            exclude_zero=exclude_zero,
            class_map=self.data["3lc_class_to_range"],
        )

    def get_validator(self, dataloader=None):
        self.loss_names = ["loss"]
        dataloader = dataloader or self.test_loader
        return TLCClassificationValidator(
            dataloader,
            self.save_dir,
            _callbacks=self.callbacks,
            run=self._run,
            image_column_name=self._image_column_name,
            label_column_name=self._label_column_name,
            settings=self._settings,
            training=True,
        )

    def get_dataloader(self, dataset_path, batch_size=16, rank=0, mode="train"):
        sampler = create_sampler(dataset_path, mode, self._settings, distributed=rank != -1)

        # Patch parent class module to use our build_dataloader
        trainer_build_dataloader = ultralytics.models.yolo.classify.train.build_dataloader
        ultralytics.models.yolo.classify.train.build_dataloader = partial(build_dataloader, sampler=sampler)

        dataloader = super().get_dataloader(dataset_path, batch_size, rank, mode)

        ultralytics.models.yolo.classify.train.build_dataloader = trainer_build_dataloader

        return dataloader



import tlc
from ultralytics.utils import colorstr

# Column names
TRAINING_PHASE = "Training Phase"
IMAGE_COLUMN_NAME = tlc.IMAGE
CLASSIFY_LABEL_COLUMN_NAME = tlc.LABEL
DETECTION_LABEL_COLUMN_NAME = f"{tlc.BOUNDING_BOXES}.{tlc.BOUNDING_BOX_LIST}.{tlc.LABEL}"
SEGMENTATION_LABEL_COLUMN_NAME = f"{tlc.SEGMENTATIONS}.{tlc.INSTANCE_PROPERTIES}.{tlc.LABEL}"
PRECISION = "precision"
PRECISION_SEG = "precision_seg"
RECALL = "recall"
RECALL_SEG = "recall_seg"
MAP = "mAP"
MAP_SEG = "mAP_seg"
MAP50_95 = "mAP50-95"
MAP50_95_SEG = "mAP50-95_seg"
NUM_IMAGES = "num_images"
NUM_INSTANCES = "num_instances"
PER_CLASS_METRICS_STREAM_NAME = "per_class_metrics"

# Other
DEFAULT_TRAIN_RUN_DESCRIPTION = ""
DEFAULT_COLLECT_RUN_DESCRIPTION = "Created with model.collect()"

TLC_PREFIX = "3LC://"
TLC_COLORSTR = colorstr("3lc: ")

REQUIREMENTS_TO_CHECK = [
    ("3lc", "tlc"),
    ("ultralytics", "ultralytics"),
]



from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path

import tlc
from ultralytics.data.utils import IMG_FORMATS, check_cls_dataset
from ultralytics.utils import ROOT, YAML

from tlc_ultralytics.utils import check_tlc_dataset


def tlc_check_cls_dataset(
    data: str,
    tables: dict[str, tlc.Table | tlc.Url | Path | str] | None,
    image_column_name: str,
    label_column_name: str,
    project_name: str | None = None,
    splits: Iterable[str] | None = None,
) -> dict[str, tlc.Table | dict[float, str] | int]:
    """Get or create tables for YOLO classification datasets. data is ignored when tables is provided.

    :param data: Path to an ImageFolder dataset
    :param tables: Dictionary of tables, if already created
    :param image_column_name: Name of the column containing image paths
    :param label_column_name: Name of the column containing labels
    :param project_name: Name of the project
    :param splits: List of splits to check for
    :return: Dictionary of tables and class names, with keys for each split and "names"
    """
    return check_tlc_dataset(
        data,
        tables,
        image_column_name,
        label_column_name,
        dataset_checker=check_cls_dataset,
        table_creator=get_or_create_cls_table,
        table_checker=check_cls_table,
        project_name=project_name,
        splits=splits,
    )


def get_or_create_cls_table(
    key: str,
    data_dict: dict[str, object],
    image_column_name: str,
    label_column_name: str,
    project_name: str,
    dataset_name: str,
    table_name: str,
) -> tlc.Table:
    """Get or create a classification table from a dataset dictionary.

    :param data_dict: Dictionary of dataset information
    :param project_name: Name of the project
    :param dataset_name: Name of the dataset
    :param table_name: Name of the table
    :param image_column_name: Name of the column containing image paths
    :param label_column_name: Name of the column containing labels
    :return: A tlc.Table.from_image_folder() table
    """

    # Special handling for ImageNet
    is_imagenet = (
        isinstance(data_dict["names"], dict)
        and isinstance(data_dict["names"][0], str)
        and data_dict["names"][0].startswith("n0")
    )

    if is_imagenet:
        label_overrides = YAML.load(ROOT / "cfg/datasets/ImageNet.yaml")["map"]
        label_overrides = {k: tlc.MapElement(internal_name=k, display_name=v) for k, v in label_overrides.items()}
    else:
        label_overrides = None

    return tlc.Table.from_image_folder(
        root=data_dict[key],
        image_column_name=image_column_name,
        label_column_name=label_column_name,
        project_name=project_name,
        dataset_name=dataset_name,
        table_name=table_name,
        extensions=IMG_FORMATS,
        if_exists="reuse",
        add_weight_column=True,
        description="Created with 3LC YOLO integration",
        label_overrides=label_overrides,
    )


def check_cls_table(table: tlc.Table, image_column_name: str, label_column_name: str) -> None:
    """Check that a table is compatible with the current task."""
    row_schema = table.row_schema.values

    try:
        # Check for image and label columns in schema
        assert image_column_name in row_schema, (
            f"Image column '{image_column_name}' not found in schema for Table {table.url}. "
            "Try providing your 'image_column_name' as an argument if you have a different column name."
        )
        assert label_column_name in row_schema, (
            f"Label column '{label_column_name}' not found in schema for Table {table.url}. "
            "Try providing your 'label_column_name' as an argument if you have a different column name."
        )

        # Check for desired roles
        assert row_schema[image_column_name].value.string_role == tlc.STRING_ROLE_IMAGE_URL, (
            f"Image column '{image_column_name}' must have role tlc.STRING_ROLE_IMAGE_URL={tlc.STRING_ROLE_IMAGE_URL}."
        )
        assert row_schema[label_column_name].value.number_role == tlc.LABEL, (
            f"Label column '{label_column_name}' must have role tlc.LABEL={tlc.LABEL}."
        )
    except AssertionError as e:
        raise ValueError(f"Table {table.url} is not compatible with YOLO classification.") from e



from __future__ import annotations

from pathlib import Path

import tlc
from ultralytics.engine.trainer import BaseTrainer
from ultralytics.utils import DEFAULT_CFG, LOGGER, RANK
from ultralytics.utils.metrics import smooth

from tlc_ultralytics.constants import DEFAULT_TRAIN_RUN_DESCRIPTION, TLC_COLORSTR
from tlc_ultralytics.engine.utils import _complete_label_column_name, _restore_random_state
from tlc_ultralytics.settings import Settings
from tlc_ultralytics.utils import reduce_embeddings


class TLCTrainerMixin(BaseTrainer):
    """A class extending the BaseTrainer class for training Ultralytics YOLO models with 3LC,
    which implements common 3LC-specific behavior across tasks. Use as a Mixin class for task-specific
    trainers.
    """

    def __init__(self, cfg=DEFAULT_CFG, overrides=None, _callbacks=None):
        LOGGER.info("Using 3LC Trainer 🌟")
        self._settings = overrides.pop("settings", Settings())
        self._settings.verify(training=True)

        assert "data" in overrides or "tables" in overrides, (
            "You must provide either a data path or tables to train with 3LC."
        )

        if "data" in overrides and not isinstance(overrides["data"], (str, Path)):
            msg = (
                "`data` must be a string or pathlib.Path. "
                "If you are using `tlc.Table` objects, provide `tables` instead."
            )
            raise ValueError(msg)

        self._tables = overrides.pop("tables", None)

        # Column names
        self._image_column_name = overrides.pop("image_column_name", self._default_image_column_name)
        self._label_column_name = overrides.pop("label_column_name", self._default_label_column_name)
        self._label_column_name = _complete_label_column_name(self._label_column_name, self._default_label_column_name)

        super().__init__(cfg, overrides, _callbacks)

        self._train_validator = None
        self._train_equals_val_result = None

        if RANK in {-1, 0}:
            self._metrics_collection_epochs = set(self._settings.get_metrics_collection_epochs(self.epochs))

            # Create a 3LC run
            description = (
                self._settings.run_description if self._settings.run_description else DEFAULT_TRAIN_RUN_DESCRIPTION
            )

            project_name = (
                self._settings.project_name if self._settings.project_name else self.data["train"].project_name
            )
            self._run = tlc.init(
                project_name=project_name,
                description=description,
                run_name=self._settings.run_name,
            )

            LOGGER.info(
                f"{TLC_COLORSTR}Created run named '{self._run.url.parts[-1]}' in project {self._run.project_name}."
            )

            # Log parameters to 3LC
            self._log_3lc_parameters()
            self._run.set_status_running()
            self._print_metrics_collection_epochs()

    def _log_3lc_parameters(self):
        """Log various data as parameters to the tlc.Run."""
        if "val" in self.data:
            val_url = str(self.data["val"].url)
        else:
            val_url = str(self.data["test"].url)

        parameters = {
            **vars(self.args),  # YOLO arguments
            "3LC/train_url": str(self.data.get("train").url),  # 3LC table used for training
            "3LC/val_url": val_url,  # 3LC table used for validation
            **{f"3LC/{k}": v for k, v in vars(self._settings).items()},  # 3LC settings
        }
        self._run.set_parameters(parameters)

    def _print_metrics_collection_epochs(self):
        """Print collection epochs to the console."""

        # Special message when no collection is enabled
        if self._settings.collection_disable:
            message = "No metrics collection is enabled."
        # No collection during training
        elif not self._metrics_collection_epochs:
            message = "Metrics will be collected after training only."
        # Print collection epochs
        else:
            if len(self._metrics_collection_epochs) == 1:
                epoch = str(next(iter(self._metrics_collection_epochs)))
                message = f"Metrics will be collected after training and after epoch {epoch}."
            else:
                epochs = ", ".join(str(epoch) for epoch in sorted(self._metrics_collection_epochs))
                message = f"Metrics will be collected after training and after the following epochs: {epochs}"

        LOGGER.info(f"{TLC_COLORSTR}{message}")

    def get_dataset(self):
        raise NotImplementedError("Subclasses must implement this method.")

    def build_dataset(self, table, mode="train", batch=None):
        raise NotImplementedError("Subclasses must implement this method.")

    def get_validator(self, dataloader):
        raise NotImplementedError("Subclasses must implement this method.")

    @property
    def train_validator(self):
        if RANK in {-1, 0}:
            if not self._train_validator:
                train_validator_dataloader = self.get_dataloader(
                    self.data["train"],
                    batch_size=self.batch_size if self.args.task == "obb" else self.batch_size * 2,
                    rank=-1,
                    mode="val",
                )
                self._train_validator = self.get_validator(dataloader=train_validator_dataloader)
            return self._train_validator
        else:
            return None

    def validate(self):
        """Perform validation with 3LC metrics collection, also on the training data, if applicable."""

        # Validate on the training set, unless the training and validation sets are identical
        if (
            not self._settings.collection_disable
            and not self._settings.collection_val_only
            and self.epoch + 1 in self._metrics_collection_epochs
            and not self._train_equals_val()
        ):
            with _restore_random_state():
                self.train_validator(trainer=self)

        # Validate on the validation/test set like usual
        return super().validate()

    def _train_equals_val(self):
        if self._train_equals_val_result is not None:
            return self._train_equals_val_result

        self._train_equals_val_result = self.data["train"] is (self.data.get("val") or self.data["test"])
        if self._train_equals_val_result:
            LOGGER.info(
                f"{TLC_COLORSTR}Training and validation sets are identical. "
                "Skipping duplicate validation on the training set."
            )
        return self._train_equals_val_result

    def final_eval(self):
        """Perform normal final validation with metrics collection on the val set, after first doing metrics collection
        on the train set.
        """
        if not self._settings.collection_val_only and not self._settings.collection_disable:
            if self.best.exists() and not self._train_equals_val():
                with _restore_random_state():
                    self.train_validator._final_validation = True
                    self.train_validator._epoch = self.epoch
                    self.train_validator.data = self.data
                    self.train_validator(model=self.best)

        self.validator._final_validation = True
        super().final_eval()
        self._save_confidence_metrics()

        if RANK in {-1, 0}:
            if self._settings.image_embeddings_dim > 0:
                train_url = self.data["train"].url
                val_url = self.data["val"].url if "val" in self.data else self.data["test"].url
                foreign_table_url = train_url if not self._settings.collection_val_only else val_url
                reduce_embeddings(
                    self._run,
                    method=self._settings.image_embeddings_reducer,
                    n_components=self._settings.image_embeddings_dim,
                    foreign_table_url=foreign_table_url,
                    reducer_args=self._settings.image_embeddings_reducer_args,
                )
            self._run.set_status_completed()

    def _save_confidence_metrics(self):
        if self.args.task not in ("detect", "segment"):
            return

        try:
            curves = [
                self.validator.metrics.box.f1_curve,  # (nc, 1000)
                self.validator.metrics.box.r_curve,  # (nc, 1000)
                self.validator.metrics.box.p_curve,  # (nc, 1000)
            ]
            names = ["F1_score", "Recall", "Precision"]
            px = self.validator.metrics.box.px  # (1000,) (linspace(0, 1)

            values = {}
            for py, name in zip(curves, names):
                y = smooth(py.mean(0), 0.05)
                best_val = y.max()
                best_conf = px[y.argmax()]
                values[f"3LC/{name}"] = {"best_val": best_val, "best_conf": best_conf}

            self._run.set_parameters(values)
        except Exception as e:
            LOGGER.error(TLC_COLORSTR + f"Failed to save confidence metrics: {e}")

    def save_metrics(self, metrics):
        # Log aggregate metrics after every epoch
        processed_metrics = self._process_metrics(metrics)

        self._run.add_output_value({"epoch": self.epoch + 1, **processed_metrics})

        super().save_metrics(metrics=metrics)

    def _process_metrics(self, metrics):
        return metrics



from __future__ import annotations

import weakref

import tlc
import torch
from ultralytics.models import yolo

from tlc_ultralytics.classify.dataset import TLCClassificationDataset
from tlc_ultralytics.classify.utils import tlc_check_cls_dataset
from tlc_ultralytics.constants import (
    CLASSIFY_LABEL_COLUMN_NAME,
    IMAGE_COLUMN_NAME,
)
from tlc_ultralytics.engine.validator import TLCValidatorMixin


class TLCClassificationValidator(TLCValidatorMixin, yolo.classify.ClassificationValidator):
    _default_image_column_name = IMAGE_COLUMN_NAME
    _default_label_column_name = CLASSIFY_LABEL_COLUMN_NAME

    def check_dataset(self, *args, **kwargs):
        return tlc_check_cls_dataset(*args, **kwargs)

    def build_dataset(self, table):
        return TLCClassificationDataset(
            table=table,
            args=self.args,
            augment=False,
            prefix=self.args.split,
            image_column_name=self._image_column_name,
            label_column_name=self._label_column_name,
            exclude_zero=self._settings.exclude_zero_weight_collection,
            class_map=self.data["3lc_class_to_range"],
        )

    def _get_metrics_schemas(self):
        predicted_value_map = self.data["names_3lc"]
        predicted_value = tlc.Int32Value(value_map=predicted_value_map, number_role=tlc.NUMBER_ROLE_LABEL)
        predicted_schema = tlc.Schema(
            "Predicted",
            "The highest confidence class predicted by the model.",
            writable=False,
            value=predicted_value,
        )

        column_schemas = {
            "loss": tlc.Schema("Loss", "Cross Entropy Loss", writable=False, value=tlc.Float32Value()),
            "predicted": predicted_schema,
            "confidence": tlc.Schema(
                "Confidence",
                "The confidence of the prediction",
                value=tlc.Float32Value(value_min=0.0, value_max=1.0),
            ),
            "top1_accuracy": tlc.Schema(
                "Top-1 Accuracy",
                "The correctness of the prediction",
                value=tlc.Float32Value(),
            ),
        }

        if len(predicted_value_map) > 5:
            column_schemas["top5_accuracy"] = tlc.Schema(
                "Top-5 Accuracy",
                "The correctness of any of the top five confidence predictions",
                value=tlc.Float32Value(),
            )

        return column_schemas

    def _compute_3lc_metrics(self, preds, batch):
        """Compute 3LC classification metrics for each sample."""
        confidence, predicted = preds.max(dim=1)

        batch_metrics = {
            "loss": torch.nn.functional.nll_loss(
                torch.log(preds), batch["cls"], reduction="none"
            ),  # nll since preds are normalized
            "predicted": [self.data["range_to_3lc_class"][int(p)] for p in predicted.tolist()],
            "confidence": confidence,
            "top1_accuracy": (torch.argmax(preds, dim=1) == batch["cls"]).to(torch.float32),
        }

        if len(self.dataloader.dataset.table.get_value_map(self._label_column_name)) > 5:
            _, top5_pred = torch.topk(preds, 5, dim=1)
            labels_expanded = batch["cls"].view(-1, 1).expand_as(top5_pred)
            top5_correct = torch.any(top5_pred == labels_expanded, dim=1)
            batch_metrics["top5_accuracy"] = top5_correct.to(torch.float32)

        return batch_metrics

    def _add_embeddings_hook(self, model):
        """Add a hook to extract embeddings from the model, and infer the activation size. For a classification model,
        this amounts to finding the linear layer and extracting the input size."""

        # Find index of the linear layer
        linear_layer_index: int | None = None
        for index, module in enumerate(model.modules()):
            if isinstance(module, torch.nn.Linear):
                activation_size = module.in_features
                linear_layer_index = index
                break

        if linear_layer_index is None:
            raise ValueError("No linear layer found in model, cannot collect embeddings.")

        weak_self = weakref.ref(self)  # Avoid circular reference (self <-> hook_fn)

        def hook_fn(_module, _input, output):
            # Store embeddings
            self_ref = weak_self()
            embeddings = output.detach().cpu().numpy()
            self_ref.embeddings = embeddings

        # Add forward hook to collect embeddings
        for i, module in enumerate(model.modules()):
            if i == linear_layer_index - 1:
                self._hook_handles.append(module.register_forward_hook(hook_fn))

        return activation_size

    def _infer_batch_size(self, preds, batch) -> int:
        return preds.size(0)



from __future__ import annotations

import importlib
import os
from dataclasses import dataclass, field, fields
from difflib import get_close_matches
from typing import Any, Callable

import tlc
from ultralytics.utils import LOGGER

from tlc_ultralytics.constants import TLC_COLORSTR


@dataclass
class Settings:
    """Settings dataclass for the 3LC integration.

    Defines and handles user settings for the 3LC integration.
    Supports parsing values from environment variables.
    """

    conf_thres: float = field(default=0.1)
    """Confidence threshold for detections. Default: 0.1"""

    max_det: int = field(default=300)
    """Maximum number of detections collected per image. Default: 300"""

    project_name: str | None = field(default=None)
    """The name of the 3LC project. Default: None"""

    run_name: str | None = field(default=None)
    """The name of the 3LC run. Default: None"""

    run_description: str | None = field(default=None)
    """The description of the 3LC run. Default: None"""

    collect_loss: bool = field(default=False)
    """Whether to collect per-sample loss values for the 'detect' task.

    Cross-Entropy loss is always computed for the 'classify' task.
    Not yet supported for the 'segment' task.

    Default: False"""

    image_embeddings_dim: int = field(default=0)
    """Image embeddings dimension. 0 means no embeddings, 2 means 2D embeddings, 3 means 3D embeddings. Default: 0"""

    image_embeddings_reducer: str = field(default="pacmap")
    """Reduction algorithm for image embeddings. Options: pacmap and umap. Only used if IMAGE_EMBEDDINGS_DIM > 0.
    Default: 'pacmap'"""

    image_embeddings_reducer_args: dict = field(default_factory=dict)
    """Reduction-method specific arguments to exert fine-grained control over the reduction process.

    See [PaCMAPTableArgs](https://docs.3lc.ai/3lc/latest/apidocs/tlc/tlc.client.reduce.pacmap.html#tlc.client.reduce.pacmap.PaCMAPTableArgs)
    or [UMAPTableArgs](https://docs.3lc.ai/3lc/latest/apidocs/tlc/tlc.client.reduce.umap.html#tlc.client.reduce.umap.UMAPTableArgs)
     for more details.
    """

    sampling_weights: bool = field(default=False)
    """Whether to use 3LC Sampling Weights. Default: False"""

    exclude_zero_weight_training: bool = field(default=False)
    """Whether to exclude zero-weighted samples in training. Default: False"""

    exclude_zero_weight_collection: bool = field(default=False)
    """Whether to exclude zero-weighted samples in metrics collection. Default: False"""

    collection_val_only: bool = field(default=False)
    """Whether to collect metrics only on the val set. Default: False"""

    collection_disable: bool = field(default=False)
    """Whether to disable 3LC metrics collection entirely. Default: False"""

    collection_epoch_start: int | None = field(default=None)
    """Start epoch for collection during training (1 is after the first epoch). Default: None"""

    collection_epoch_interval: int = field(default=1)
    """Epoch interval for collection. Only used if a starting epoch is set. Default: 1"""

    metrics_collection_function: Callable[[Any, Any], dict[str, Any]] | None = field(default=None)
    """Function to compute additional metrics during collection.

    This function should take predictions and batch data as arguments and return
    a dictionary of additional metrics to be included in the collection.

    Args:
        preds: Model predictions (format depends on task)
        batch: Input batch data

    Returns:
        Dictionary of additional metrics to collect

    Example:
        def custom_metrics(preds, batch):
            return {"custom_metric": compute_something(preds, batch)}

    Note:
        This function must be pickleable if using multiprocessing. Avoid lambda
        functions and ensure the function is defined at module level.

    Default: None"""

    metrics_schemas: dict[str, tlc.Schema] | None = field(default=None)
    """Schemas for any additional metrics returned by the metrics_collection_function.

    Providing schemas is optional, but for complex metrics, it is recommended to provide a schema.

    Default: None"""

    @classmethod
    def from_env(cls) -> Settings:
        """Create a Settings instance from environment variables.

        :returns: A Settings instance with values parsed from environment variables.
        """
        cls._handle_unsupported_env_vars()  # Warn about unsupported environment variables

        kwargs = {}

        for _field in fields(cls):
            env_var = cls._field_to_env_var(_field)
            env_value = os.getenv(env_var, None)
            if env_value is not None:
                value = Settings._parse_env_var(env_var, env_value, _field.type)
                kwargs[_field.name] = value

        instance = cls(**kwargs)

        instance._from_env = True  # Mark the instance as created from environment variables

        return instance

    def __post_init__(self) -> None:
        # Mark as not created from environment variables
        if not hasattr(self, "_from_env"):
            self._from_env = False

    def verify(self, training: bool = True) -> None:
        """Verify that the settings are valid.

        :param training: Whether the settings are for training or validation.

        :raises: AssertionError if the settings are invalid.
        """
        # Checks
        assert 0.0 <= self.conf_thres <= 1.0, f"Confidence threshold {self.conf_thres} is not in [0, 1]."
        assert self.max_det > 0, f"Maximum number of detections {self.max_det} is not positive."
        assert self.image_embeddings_dim in (0, 2, 3), (
            f"Invalid image embeddings dimension {self.image_embeddings_dim}."
        )
        if self.image_embeddings_dim > 0:
            self._check_reducer_available()

        # Validate metrics collection function if provided
        if self.metrics_collection_function is not None:
            assert callable(self.metrics_collection_function), (
                f"metrics_collection_function must be callable, got {type(self.metrics_collection_function)}"
            )

        # Train / collect specific settings
        self._verify_training() if training else self._verify_collection()

    def get_metrics_collection_epochs(self, epochs: int) -> list[int]:
        if self.collection_disable:
            return []

        if self.collection_epoch_start is None:
            return []

        if self.collection_epoch_start > epochs:
            return []

        # If start is less than one, we don't collect during training
        if self.collection_epoch_start < 1:
            raise ValueError(
                f"Invalid collection start epoch {self.collection_epoch_start}, must be at least 1 (after first epoch)."
            )

        if self.collection_epoch_interval <= 0:
            raise ValueError(f"Invalid interval {self.collection_epoch_interval}, must be non-zero")
        else:
            return list(
                range(
                    self.collection_epoch_start,
                    epochs + 1,
                    self.collection_epoch_interval,
                )
            )

    def _verify_training(self) -> None:
        """Verify that the settings are valid for training.

        :param opt: The argparse namespace containing YOLO settings.

        :raises: AssertionError if the settings are invalid.
        """
        # Can't collect things when collection is disabled
        cases = [
            (self.collection_val_only, "collect only on val set"),
            # (self.collect_loss, 'collect loss values'), TODO: Restore when loss is supported.
            (self.image_embeddings_dim > 0, "collect image embeddings"),
            (self.collection_epoch_start, "collect metrics during training"),
        ]

        for setting, description in cases:
            assert not (self.collection_disable and setting), f"Cannot disable collection and {description}."

        # Collection epoch settings
        assert self.collection_epoch_start is None or self.collection_epoch_start >= 0, (
            f"Invalid collection start epoch {self.collection_epoch_start}."
        )
        assert self.collection_epoch_interval > 0, (
            f"Invalid collection epoch interval {self.collection_epoch_interval}."
        )

    def _verify_collection(self) -> None:
        """Verify that the settings are valid for metrics collection only (no training).

        :raises: AssertionError if the settings are invalid.
        """
        pass

    def _check_reducer_available(self) -> None:
        """Check that the selected reducer is available.

        :raises: ValueError if the selected reducer is not available.
        """
        reducer_to_package = {"pacmap": "pacmap", "umap": "umap-learn"}
        if self.image_embeddings_reducer not in reducer_to_package:
            raise ValueError(
                f"Invalid image embeddings reducer {self.image_embeddings_reducer}. "
                "Valid options are 'pacmap' and 'umap'."
            )

        try:
            importlib.import_module(self.image_embeddings_reducer)
        except Exception as e:
            package = reducer_to_package[self.image_embeddings_reducer]
            raise ValueError(
                f"Embeddings collection enabled, but failed to import {self.image_embeddings_reducer} dependency. "
                f"Run `pip install {package}` to enable embeddings collection."
            ) from e

    @staticmethod
    def _field_to_env_var(_field: field) -> None:
        """Return the environment variable name for a given field.

        :param _field: The field to get the environment variable for.
        :returns: The environment variable name.
        """
        return f"TLC_{_field.name.upper()}"

    def _handle_unsupported_env_vars(self) -> None:
        """Handle environment variables starting with TLC which are not supported.

        Appropriate warnings are logged when unsupported environment variables are encountered.
        """
        supported_env_vars = [self._field_to_env_var(_field) for _field in fields(Settings)]
        unsupported_env_vars = [var for var in os.environ if var.startswith("TLC_") and var not in supported_env_vars]

        # Output all environment variables if there are any unsupported ones
        if len(unsupported_env_vars) > 1:
            LOGGER.warning(
                f"{TLC_COLORSTR}Found unsupported environment variables: "
                f"{', '.join(unsupported_env_vars)}.\n{self._supported_env_vars_str()}"
            )

        # If there is only one, look for the most similar one
        elif len(unsupported_env_vars) == 1:
            closest_match = get_close_matches(unsupported_env_vars[0], supported_env_vars, n=1, cutoff=0.4)
            if closest_match:
                LOGGER.warning(
                    f"{TLC_COLORSTR}Found unsupported environment variable: {unsupported_env_vars[0]}. "
                    f"Did you mean {closest_match[0]}?"
                )
            else:
                LOGGER.warning(
                    f"{TLC_COLORSTR}Found unsupported environment variable: {unsupported_env_vars[0]}."
                    f"\n{self._supported_env_vars_str()}"
                )

    def _supported_env_vars_str(self, sep: str = "\n  - ") -> str:
        """Print all supported environment variables.

        :param sep: The separator to use between each variable.
        :returns: A sep-separated string with all supported environment variables.

        """
        default_settings_instance = Settings()  # Create an instance to get the default values

        # Display defaults differently for environment variables as they are provided differently
        if self._from_env:

            def formatter(x):
                return x if not isinstance(x, list) else ",".join(x)
        else:

            def formatter(x):
                return x

        field_info_list = [
            {
                "name": self._field_to_env_var(_field),
                "description": _field.metadata["description"],
                "default": formatter(getattr(default_settings_instance, _field.name)),
            }
            for _field in fields(Settings)
        ]

        # Display each line as TLC_<FIELD_NAME>: <DESCRIPTION>. Default: '<DEFAULT>'
        lines = [
            f"{field_info['name']}: {field_info['description']}. Default: '{field_info['default']}'."
            for field_info in field_info_list
        ]
        return f"Supported environment variables:{sep}{sep.join(lines)}"

    @staticmethod
    def _parse_env_var(name: str, value: Any, var_type: str) -> Any:
        """Parse an environment variable.

        :param name: The name of the environment variable.
        :param value: The value of the environment variable.
        :param var_type: The expected type of the environment variable as defined in the dataclass.
        """
        if var_type == "bool":
            return Settings._parse_boolean_env_var(name, value)
        elif var_type == "list":
            return value.split(",")
        elif var_type == "int":
            return int(value)
        elif var_type == "float":
            return float(value)
        elif var_type == "str":
            return value
        else:
            raise ValueError(f"Unsupported type {var_type} for environment variable {name}.")

    @staticmethod
    def _parse_boolean_env_var(name: str, value: Any) -> bool:
        """Parse a boolean environment variable. Supported values:
        - true/false (case insensitive)
        - y/n (case insensitive)
        - 1/0
        - yes/no (case insensitive)

        :param name: The name of the environment variable.
        :param default: The value of the environment variable.
        :returns: The parsed boolean value.
        :raises: ValueError if the value is not a valid boolean.
        """
        if value.lower() in ("y", "yes", "1", "true"):
            return True
        elif value.lower() in ("n", "no", "0", "false"):
            return False
        else:
            raise ValueError(
                f"Invalid value {value} for environment variable {name}, "
                "should be a boolean on the form y/n, yes/no, 1/0 or true/false."
            )



from .check_requirements import check_requirements
from .dataset import check_tlc_dataset, parse_3lc_yaml_file
from .embeddings import reduce_embeddings
from .sampler import create_sampler
from .schemas import image_embeddings_schema, training_phase_schema

__all__ = (
    "check_requirements",
    "check_tlc_dataset",
    "create_sampler",
    "get_table_value_map",
    "image_embeddings_schema",
    "parse_3lc_yaml_file",
    "reduce_embeddings",
    "training_phase_schema",
)



import contextlib
import random


def _complete_label_column_name(label_column_name: str, default_label_column_name: str) -> str:
    """Create a complete label column name from a potentially partial one.

    Examples:
        >>> _complete_label_column_name("a", "a")
        "a"
        >>> _complete_label_column_name("a", "a.b.c")
        "a.b.c"
        >>> _complete_label_column_name("a.b.c", "d.e.f")
        "a.b.c"
        >>> _complete_label_column_name("", "a.b.c")
        "a.b.c"
    """
    parts = label_column_name.split(".") if label_column_name else []
    default_parts = default_label_column_name.split(".")

    for i, default_part in enumerate(default_parts):
        if i >= len(parts):
            parts.append(default_part)

    return ".".join(parts)


@contextlib.contextmanager
def _restore_random_state():
    """Context manager to ensure the global random state is unchanged by the wrapped code."""
    state = random.getstate()
    yield
    random.setstate(state)



from __future__ import annotations

from typing import Literal

import tlc
import torch

from tlc_ultralytics.settings import Settings


def create_sampler(
    table: tlc.Table,
    mode: Literal["train", "val"],
    settings: Settings,
    distributed: bool = False,
) -> torch.utils.data.Sampler | None:
    """Get the sampler for the dataset.

    :param table: The table to get the sampler for.
    :param mode: The mode of the sampler.
    :param settings: The settings for the run.
    :param distributed: Whether training is distributed.
    :returns: The sampler for the dataset.
    """
    sampler = None

    if mode == "train":
        if settings.sampling_weights or settings.exclude_zero_weight_training:
            if distributed:
                raise NotImplementedError("Distributed training and using 3LC weights is not yet supported.")

            # No need to exclude zero weights if there is no weights column
            exclude_zero_weights = False if table.weights_column_name is None else settings.exclude_zero_weight_training

            try:
                sampler = table.create_sampler(
                    exclude_zero_weights=exclude_zero_weights,
                    weighted=settings.sampling_weights,
                    shuffle=True,
                )
            except Exception as e:
                raise ValueError(f"Error creating sampler for table {table.url}") from e

    elif mode == "val":
        if distributed:
            raise NotImplementedError("Distributed validation and exclusion by weight is not yet supported.")

        # Exclude zero weight is handled in the dataset for validation
        return None
    return sampler



from __future__ import annotations

from functools import partial

import ultralytics
from ultralytics.models.yolo.detect import DetectionTrainer

from tlc_ultralytics.constants import (
    DETECTION_LABEL_COLUMN_NAME,
    IMAGE_COLUMN_NAME,
)
from tlc_ultralytics.detect.utils import (
    build_tlc_yolo_dataset,
    tlc_check_det_dataset,
)
from tlc_ultralytics.detect.validator import TLCDetectionValidator
from tlc_ultralytics.engine.trainer import TLCTrainerMixin
from tlc_ultralytics.overrides import build_dataloader
from tlc_ultralytics.utils import create_sampler


class TLCDetectionTrainer(TLCTrainerMixin, DetectionTrainer):
    """Trainer class for YOLO object detection with 3LC"""

    _default_image_column_name = IMAGE_COLUMN_NAME
    _default_label_column_name = DETECTION_LABEL_COLUMN_NAME

    def get_dataset(self):
        # Parse yaml and create tables
        self.data = tlc_check_det_dataset(
            self.args.data,
            self._tables,
            self._image_column_name,
            self._label_column_name,
            project_name=self._settings.project_name,
            splits=("train", "val"),
        )

        # Get test data if val not present
        if "val" not in self.data:
            data_test = tlc_check_det_dataset(
                self.args.data,
                self._tables,
                self._image_column_name,
                self._label_column_name,
                project_name=self._settings.project_name,
                splits=("test",),
            )
            self.data["test"] = data_test["test"]

        return self.data

    def build_dataset(self, *args, **kwargs):
        from ultralytics.models.yolo.detect.train import build_yolo_dataset as original_build_yolo_dataset

        mode = kwargs.get("mode") or args[1]

        exclude_zero = mode == "val" and self._settings.exclude_zero_weight_collection
        ultralytics.models.yolo.detect.train.build_yolo_dataset = partial(
            build_tlc_yolo_dataset,
            exclude_zero=exclude_zero,
            class_map=self.data["3lc_class_to_range"],
            image_column_name=self._image_column_name,
            label_column_name=self._label_column_name,
        )

        result = DetectionTrainer.build_dataset(self, *args, **kwargs)

        ultralytics.models.yolo.detect.train.build_yolo_dataset = original_build_yolo_dataset

        return result

    def get_validator(self, dataloader=None):
        self.loss_names = "box_loss", "cls_loss", "dfl_loss"
        if not dataloader:
            dataloader = self.test_loader

        return TLCDetectionValidator(
            dataloader,
            save_dir=self.save_dir,
            args=self.args,
            run=self._run,
            image_column_name=self._image_column_name,
            label_column_name=self._label_column_name,
            settings=self._settings,
            training=True,
        )

    def _process_metrics(self, metrics):
        return {
            metric.removesuffix("(B)").replace("metrics", "val").replace("/", "_"): value
            for metric, value in metrics.items()
        }

    def get_dataloader(self, dataset_path, batch_size=16, rank=0, mode="train"):
        """Construct and return dataloader."""

        sampler = create_sampler(dataset_path, mode, self._settings, distributed=rank != -1)

        # Patch parent class module to use our build_dataloader
        trainer_build_dataloader = ultralytics.models.yolo.detect.train.build_dataloader
        ultralytics.models.yolo.detect.train.build_dataloader = partial(build_dataloader, sampler=sampler)

        dataloader = super().get_dataloader(dataset_path, batch_size, rank, mode)

        # Restore parent class module
        ultralytics.models.yolo.detect.train.build_dataloader = trainer_build_dataloader

        return dataloader
